//------------------------------------------------------------------------------
//The contents of this file are subject to the Mozilla Public License
//Version 1.1 (the "License"); you may not use this file except in compliance
//with the License. You may obtain a copy of the License at
//http://www.mozilla.org/MPL/ Software distributed under the License is
//distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
//or implied. See the License for the specific language governing rights and
//limitations under the License.
//
//The Original Code is AsmLst.pas.
//
//The Initial Developer of the Original Code is Alex Shovkoplyas, VE3NEA.
//Portions created by Alex Shovkoplyas are
//Copyright (C) 2008 Alex Shovkoplyas. All Rights Reserved.
//------------------------------------------------------------------------------
unit AsmLst;

interface

uses
  Windows, SysUtils, Classes, ModulInf, JclPeImage, JclDebug, Math,
  DbgSrc, FileFind, NdisAsm, JclWin32;


const
  BUF_SIZE = 4096;
  HALF_BUF_SIZE = BUF_SIZE div 2;
  QUARTER_BUF_SIZE = BUF_SIZE div 4;


type
  TIntegerArray = array of Integer;


  TGetSourceCodeEvent = procedure(Sender: TObject; AModule: TModuleItem;
    AUnitName: string; ALines: TStringList) of object;

  TGetSyncPointsEvent = procedure(Sender: TObject; AStartAddr: DWord;
    APoints: TIntegerArray) of object;


  TAsmLister = class
  private
    FModules: TModuleList;
    FModule: TModuleItem;
    FImage: TJclPeImage;
    FSect: PImageSectionHeader;
    FSectName: string;
    FComments: array of string;
    FSrcUnitName: TFileName;
    FSrcLines: TStringList;
    DataArr, Relocs: array of Byte;

    //buffer start and end, in terms of runtime addresses
    FAddrB, FAddrE: DWord;
    //buffer start and end relative to imagebase
    FRvaB, FRvaE: DWord;
    //section start and end, in terms of runtime addresses
    FSectB, FSectE: DWord;
    //pointer to data corresponding to FAddrB
    FData: PByteArray;
    FOnGetSourceCode: TGetSourceCodeEvent;
    FOnGetSyncPoints: TGetSyncPointsEvent;

    function GetBinFile: TFileName;
    procedure SetupBuffer(AAddr: DWord);
    function IsBufOK(AAddr: DWord): boolean;
    procedure Disassemble;
    procedure GetSectionFromAddr(AAddr: DWord);
    procedure GetComments;
    function GetSourceLine(AUnit: string; ALineNo: integer): string;
    procedure Relocate;
    procedure PrintDB(var Idx: integer);
    function IsPossibleDD(Idx: integer): boolean;
    function OperandsToSymbols(S: string): string;
    function AddrToSymbol(Addr: DWord): string;
    procedure InsertComments;
  public
    Lines: TStringList;
    LineAtAddr: integer;
    Addresses: array of DWord;
    ErrText: string;

    constructor Create(AModules: TModuleList);
    destructor Destroy; override;
    procedure Clear;
    function MakeLines(AAddr: DWord; ACount: integer): boolean;

    property BinFile: TFileName read GetBinFile;
    property OnGetSourceCode: TGetSourceCodeEvent read FOnGetSourceCode write FOnGetSourceCode;
    property OnGetSyncPoints: TGetSyncPointsEvent read FOnGetSyncPoints write FOnGetSyncPoints;
  end;


implementation


{ TAsmLister }

//------------------------------------------------------------------------------
//                                 system
//------------------------------------------------------------------------------
constructor TAsmLister.Create;
begin
  FModules := AModules;
  Lines := TStringList.Create;
  FImage := TJclPeImage.Create(True);
  FSrcLines := TStringList.Create;
end;


destructor TAsmLister.Destroy;
begin
  Lines.Free;
  FImage.Free;
  FSrcLines.Free;
  inherited;
end;


procedure TAsmLister.Clear;
begin
  Lines.Clear;
  FModule := nil;
  FSect := nil;
  FData := nil;
  FAddrB := 0;
  FAddrE := 0;
end;






//------------------------------------------------------------------------------
//                        get bin data to disassemble
//------------------------------------------------------------------------------
function TAsmLister.GetBinFile: TFileName;
begin
  if FModule = nil then Result := '' else Result := FModule.BinFile;
end;


function TAsmLister.MakeLines(AAddr: DWord; ACount: integer): boolean;
var
  i: integer;
begin
  Result := false;
  ErrText := '';

  //ensure module
  if (FModule = nil) or not FModule.IsAddressInModule(AAddr) then
    begin
    //find module from addr
    FModule := FModules[FModules.GetModuleIndexFromAddr(AAddr)];
    if FModule.BinFile = STR_UNKNOWN then
      begin ErrText := 'Unknown module'; Exit; end;

    //load module
    if not FModule.BinExists then
      begin ErrText := 'File does not exist'; Exit; end;
    FImage.FileName := BinFile;
    Result := FImage.StatusOK;
    if not Result then
      begin ErrText := 'Unable to load file'; Exit; end;

    //module changed, discard old section
    FSect := nil;
    end;

  //ensure section
  if (FSect = nil) or (AAddr < FSectB) or (AAddr >= FSectE) then
    begin
    //find section
    GetSectionFromAddr(AAddr);
    if FSect = nil then
      begin ErrText := 'Address outside of the code section'; Exit; end;

    //section changed, discard old buffer
    FData := nil;
    end;

  //ensure buffer, disasm if changed
  if not IsBufOK(AAddr) then
    begin
    SetupBuffer(AAddr);
    GetComments;
    Disassemble;
    InsertComments;
    Result := true;
    end;

  //find line in listing
  for i:=0 to High(Addresses) do
    if Addresses[i] = AAddr then
      begin LineAtAddr := i; Break; end
    else if Addresses[i] > AAddr then
      begin LineAtAddr := i - 1; Break; end;
end;


procedure TAsmLister.GetSectionFromAddr(AAddr: DWord);
var
  i: integer;
begin
  for i:=0 to FImage.ImageSectionCount - 1 do
    begin
    //get section header
    if not FImage.GetSectionHeader(FImage.ImageSectionNames[i], FSect)
      then Continue;

    //section start/end addr at runtime
    FSectB := FModule.BaseAddr + FSect.VirtualAddress;
    FSectE := FSectB + FSect.Misc.VirtualSize;
    SetString(FSectName, PChar(@FSect.Name[0]), 8);
    FSectName := Trim(FSectName);

    //AAddr in section
    if (AAddr >= FSectB) and (AAddr < FSectE) then Exit;
    end;

  //not found
  FSect := nil;
end;


function TAsmLister.IsBufOK(AAddr: DWord): boolean;
begin
  //buffer not available
  Result := false;
  if FData = nil then Exit;

  //AAddr not in buffer
  Result := (AAddr >= FAddrB) and (AAddr < FAddrE);
  if not Result then Exit;

  //AAddr must be in the middle of the buffer,
  //unless near section edge
  Result := (AAddr >= (FAddrB + QUARTER_BUF_SIZE)) or (FAddrB = FSectB);
  if not Result then Exit;

  Result := (AAddr < (FAddrE - QUARTER_BUF_SIZE)) or (FAddrE = FSectE);
end;


procedure TAsmLister.SetupBuffer(AAddr: DWord);
var
  A: DWord;
begin
  FAddrB := AAddr and not (HALF_BUF_SIZE - 1);
  if (AAddr < (FAddrB + QUARTER_BUF_SIZE)) and (FAddrB > FSectB) then
    Dec(FAddrB, HALF_BUF_SIZE);

  FAddrE := FAddrB + BUF_SIZE;
  if FAddrE > FSectE then FAddrE := FSectE;

  FRvaB := FAddrB - FModule.BaseAddr;
  FRvaE := FAddrE - FModule.BaseAddr;

  //offset from section start
  A := FRvaB - FSect.VirtualAddress;
  //offset in exe file
  A := FSect.PointerToRawData + A;
  //pointer to data in buffer                         
  FData := Pointer(FImage.LoadedImage.MappedAddress + A);
end;


procedure TAsmLister.Relocate;
var
  Delta: integer;
  i, j, Idx: integer;
begin
  //actual image base vs. default base
  Delta := Integer(FModule.BaseAddr) - Integer(FImage.OptionalHeader32.ImageBase);

  //buffer for relocation points
  Relocs := nil;
  SetLength(Relocs, FAddrE - FAddrB + 32);

  //make a copy of data
  DataArr := nil;
  SetLength(DataArr, FAddrE - FAddrB + 32);
  Move(FData^, DataArr[0], FAddrE - FAddrB);
  FData := @DataArr[0];

  //find item in the relocation table
  for i:=0 to FImage.RelocationList.Count-1 do
    if FImage.RelocationList.Items[i].VirtualAddress < (FRvaB and $FFFFF000)
      then Continue
    else if FImage.RelocationList.Items[i].VirtualAddress >= FRvaE
      then Break
    else
      //apply relocations
      for j:=0 to FImage.RelocationList.Items[i].Count-1 do
        with FImage.RelocationList.Items[i].Relocations[j] do
          if (RelocType = IMAGE_REL_BASED_HIGHLOW) and (VirtualAddress >= FRvaB) and (VirtualAddress < FRvaE) then
            begin
            Idx := VirtualAddress - FRvaB;
            Inc(PInteger(@DataArr[Idx])^, Delta);
            Relocs[Idx] := Byte(true);
            end;
end;






//------------------------------------------------------------------------------
//                               comments
//------------------------------------------------------------------------------
procedure TAsmLister.GetComments;
var
  Idx: integer;
  LocInfo: TJclLocationInfo;
  S: string;
begin
  FComments := nil;
  SetLength(FComments, FAddrE - FAddrB);
  Idx := High(FComments);

  repeat
    FModule.DebugInfoSource.GetLocationInfo(Ptr(FAddrB + Idx), LocInfo);

    //show symbols as labels
    if (LocInfo.ProcedureName <> STR_UNKNOWN) and (LocInfo.ProcedureName <> '') then
      if LocInfo.OffsetFromProcName = 0 then
        begin
        S := LocInfo.ProcedureName;
        if LocInfo.UnitName <> STR_UNKNOWN then S := LocInfo.UnitName + '.' + S;
        FComments[Idx] := Format('''%s''', [S]);
        end;

    //insert source code lines as comments
    if LocInfo.LineNumber > 0 then
      if LocInfo.OffsetFromLineNumber = 0 then
        begin
        S := Trim(GetSourceLine(LocInfo.UnitName, LocInfo.LineNumber));
        if S <> '' then S := ':  ' + S;
        FComments[Idx] := FComments[Idx] + Format(';%s.%d', [LocInfo.UnitName, LocInfo.LineNumber]) + S;
        end;

    Dec(Idx);
  until Idx < 0;


{
  //this method misses a lot of symbols
  if FModule.DebugInfoSource.Kind <> diBorSymb then
    if FModule.DebugInfoSource.GetLocationInfo(Ptr(FAddrB), LocInfo) then
      begin
      Addr := DWord(LocInfo.Address);
      repeat
        if not FModule.DebugInfoSource.GetNextFunc(Addr, LocInfo)
          then begin beep; Break; end;
        Addr := DWord(LocInfo.Address);
        if (Addr = 0) or (Addr >= FAddrE)
          then begin beep; Break; end;
        FComments[Addr-FAddrB] := FComments[Addr-FAddrB] + '<' + LocInfo.ProcedureName + '>';
      until false;
      end;
}
end;


procedure TAsmLister.PrintDB(var Idx: integer);
begin
  Lines.AddObject(Format('        db $%.2x', [FData[Idx]]), Ptr(FAddrB + Idx));
  Inc(Idx);
end;


function TAsmLister.IsPossibleDD(Idx: integer): boolean;
begin
 if  Idx >= (FAddrE - FAddrB)
   then Result := false
   else Result := (Relocs[Idx] <> 0) or (PInteger(@FData[Idx])^ = -1);// or (PInteger(@FData[Idx])^ = 0);
end;


procedure TAsmLister.InsertComments;
const
  SecStart = ';----Start of section [%s]-----------------------------------------------------------';
  SecEnd   = ';----End of section [%s]-------------------------------------------------------------';
var
  L, p, Idx: integer;
  S: string;
begin
  for L:= Lines.Count-1 downto 0 do
    begin
    Idx := DWord(Lines.Objects[L]) - FAddrB;

    //comment lines
    if FComments[Idx] <> '' then
      begin
      p := Pos(';', FComments[Idx]);
      if p < 2
        then
          Lines.Insert(L, FComments[Idx])
        else
          begin
          Lines.Insert(L, Copy(FComments[Idx], p, MAXINT));
          Lines.Insert(L, Copy(FComments[Idx], 1, p-1));
          end;
      //blank line before comments
      Lines.Insert(L, '');
      end;

    //blank line after jmp/ret
    if (L > 0) and (Lines[L] <> '') then
      begin
      S := Copy(Lines[L-1], 9, 4);
      if (S = 'ret ') or (S = 'iret') or (S = 'jmp ') or (S = 'halt')
        then Lines.Insert(L, '');
      end;
    end;

  //add section start comment
  if FAddrB = FSectB then
    begin
    S := Format(SecStart, [FSectName]);
    SetLength(S, 80);
    Lines.Insert(0, S);
    end;

  //add section end comment
  if FAddrE = FSectE then
    begin
    S := Format(SecEnd, [FSectName]);
    SetLength(S, 80);
    Lines.Add(S);
    end;


{
  //remove lines with ponential synch errors
  if (Lines.Count > 32) and (FAddrB > FSectB) then
    begin
    DelCnt := 32;
    if (Copy(Lines[DelCnt-1], 1, 1) = '''') or
       (Copy(Lines[DelCnt-1], 1, 1) = ';') then Dec(DelCnt);
    if (Copy(Lines[DelCnt-1], 1, 1) = '''') or
       (Copy(Lines[DelCnt-1], 1, 1) = ';') then Dec(DelCnt);
    for j:=DelCnt-1 downto 0 do Lines.Delete(j);
    end;
}


  //move line addresses to array
  //clear Lines.Objects as they will be used for hit counts
  Addresses := nil;
  SetLength(Addresses, Lines.Count);
  for L:=0 to Lines.Count-1 do
    begin
    Addresses[L] := DWord(Lines.Objects[L]);
    Lines.Objects[L] := nil;
    end;
end;


function TAsmLister.GetSourceLine(AUnit: string; ALineNo: integer): string;
begin
  //load source code
  if FSrcUnitName <> AUnit then
    begin
    if Assigned(FOnGetSourceCode) then FOnGetSourceCode(Self, FModule, AUnit, FSrcLines);
    FSrcUnitName := AUnit;
    end;

  //return line
  if ALineNo <= FSrcLines.Count
    then Result := FSrcLines[ALineNo-1] else Result := '';
end;






//------------------------------------------------------------------------------
//                               disassember
//------------------------------------------------------------------------------
procedure TAsmLister.Disassemble;
var
  i, j: integer;
  Idx: integer;
  S: string;
  SyncPoints: TIntegerArray;
  SyncErr, AddrTable: boolean;
begin
  SetLength(SyncPoints, FAddrE - FAddrB);
  if Assigned(FOnGetSyncPoints) then FOnGetSyncPoints(Self, FAddrB, SyncPoints);

  Relocate;

  Lines.Clear;
  Idx := 0;

  repeat
    DisassembleInstruction( @FData[Idx], FAddrB + Idx);

    //synch error if bad instruction
    SyncErr := NDisAsm.DisAsmLength = 0;

    //detect synch errors using symbol info and profiler hit addresses
    if not SyncErr then
      for i:=1 to Min(High(FComments)-Idx, NDisAsm.DisAsmLength-1) do
        if ((FComments[Idx+i] <> '') and (FComments[Idx+i][1] in ['''',';']))
            or (SyncPoints[Idx+i] <> 0) then
          begin SyncErr := true; Break; end;

    //if relocation at 1-st byte of instr, sync error
    if IsPossibleDD(Idx) then SyncErr := true;

    //detect address tables using relocation info
    AddrTable := false;
    for i:=0 to Max(0, NDisAsm.DisAsmLength-1) do
      //find tables
      if IsPossibleDD(Idx+i) and IsPossibleDD(Idx+i+4) then
        begin
        for j:=0 to i-1 do PrintDB(Idx);
        AddrTable := true;
        Break;
        end;

    //spit out DD if table
    if AddrTable then
      begin
      repeat
        S := Format('        dd $%.8x', [PInteger(@FData[Idx])^]);
        Lines.AddObject(OperandsToSymbols(S), Ptr(FAddrB + Idx));
        Inc(Idx, 4);
      until
        not IsPossibleDD(Idx);
      Continue;
      end

    //if sync error, spit out DB
    else if SyncErr then
      begin
      PrintDB(Idx);
      Continue;
      end;

    //make string
    S := StringReplace(NDisAsm.DisAsmString, '0x', '$', [rfReplaceAll]);
    S := OperandsToSymbols(S);
    S := '        ' + StringReplace(S, '         ', ' ', []);

    //add raw bytes
    S := S + StringOfChar(' ', 56 - Length(S));
    for j :=0 to NDisAsm.DisAsmLength-1 do S := S + ' $' + IntToHex(FData[Idx+j], 2);

    //output
    Lines.AddObject(S, Ptr(FAddrB + Idx));
    Inc(Idx, NDisAsm.DisAsmLength);
  until
    (FAddrB + Idx) >= FAddrE;

end;






//------------------------------------------------------------------------------
//                       replace addresses with symbols
//------------------------------------------------------------------------------
function TAsmLister.OperandsToSymbols(S: string): string;
const
  HexDigs = ['0'..'9','a'..'f','A'..'F'];
var
  i, IdxB, IdxE: integer;
  Hex, Symb: string;
  Addr: DWord;
  Neg: boolean;
begin
  Result := S;

  if Length(S) < 9 then Exit;

  IdxE := 4;

  repeat
    //find start of hex value
    IdxB := Length(Result) + 1;
    for i:=IdxE to Length(Result)-5 do
      if Result[i] = '$' then begin IdxB := i; Break; end;
    if IdxB > Length(Result) then Exit;

    //find end of hex value
    IdxE := Length(Result) + 1;
    for i:=IdxB+1 to Length(Result) do
      if not (Result[i] in HexDigs) then begin IdxE := i; Break; end;

    //extract value
    Hex := Copy(Result, IdxB, IdxE-IdxB);
    Addr := StrToInt64Def(Hex, 0);
    Neg := (IdxB > 1) and (Result[IdxB-1] = '-');
    if Neg then Addr := -Addr;

    //replace with symbol
    if (Addr >= FModule.BaseAddr) and (Addr < FModule.EndAddr) then
      begin
      if Neg then Result[IdxB-1] := '+';
      Symb := AddrToSymbol(Addr);
      Result := Copy(Result, 1, IdxB-1) + Symb + Copy(Result, IdxE, MAXINT);
      Inc(IdxE, Length(Symb) - Length(Hex));
      end;
  until IdxE > Length(Result);
end;      


function TAsmLister.AddrToSymbol(Addr: DWord): string;
var
  LocInfo: TJclLocationInfo;
  LabelName: string;
  OpCode: DWord;
begin
  //resolve only within the module
  if (Addr < FModule.BaseAddr) or (Addr >= FModule.EndAddr) then
    Result := Format('<$%.8x>', [Addr])

  //target within displayed page
  else if (Addr >= FAddrB) and (Addr < FAddrE) then
    begin
    Result := FComments[Addr - FAddrB];

    //no label, create
    if Result = '' then
      begin
      OpCode := NdisAsm.DisAsmInfo.Templ.opcode;
      case OpCode of
        I_CALL:
          LabelName := 'Proc';

        I_JMP,I_Jcc:
          LabelName := 'Label';

        I_PUSH:
          if (NdisAsm.DisAsmInfo.Templ.opd[0] and IMMEDIATE) <> 0
            then LabelName := 'Label' else LabelName := 'Data';

        else
          LabelName := 'Data';
        end;

      Result := Format('%s_%.3x', [LabelName, Addr and $FFF]);
      FComments[Addr - FAddrB] :=  Result + ':';
      end

    //jump label exists
    else if Result[1] in ['L','P','D'] then
      Result := Copy(Result, 1, Pos(':', Result)-1)

    //symbol label exists
    else if Result[1] = '''' then
      begin
      Delete(Result, 1, 1);
      Result := Copy(Result, 1, Pos('''', Result)-1);
      end

    //line label exists
    else if Result[1] = ';' then
      begin
      Delete(Result, 1, 1);
      Result := Copy(Result, 1, Pos(':', Result)-1);
      end;
    end

  //not in the page, request symbol info
  else if FModule.DebugInfoSource.GetLocationInfo(Ptr(Addr), LocInfo) then
    begin
    //function symbol found
    if (LocInfo.ProcedureName <> STR_UNKNOWN) and
       (LocInfo.ProcedureName <> '') and
       (LocInfo.OffsetFromProcName = 0) then
      begin
      Result := LocInfo.ProcedureName;
      if LocInfo.UnitName <> STR_UNKNOWN then Result := LocInfo.UnitName + '.' + Result;
      end

    //line number found
    else if (LocInfo.LineNumber > 0) and (LocInfo.OffsetFromLineNumber = 0) 
      then Result := Format('%s.%d', [LocInfo.UnitName, LocInfo.LineNumber])

    //no symbol at that address
    else Result := Format('<$%.8x>', [Addr]);
    end

  //no symbol found
  else Result := Format('<$%.8x>', [Addr]);
end;



end.


