//------------------------------------------------------------------------------
//The contents of this file are subject to the Mozilla Public License
//Version 1.1 (the "License"); you may not use this file except in compliance
//with the License. You may obtain a copy of the License at
//http://www.mozilla.org/MPL/ Software distributed under the License is
//distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
//or implied. See the License for the specific language governing rights and
//limitations under the License.
//
//The Original Code is BorPrj.pas.
//
//The Initial Developer of the Original Code is Alex Shovkoplyas, VE3NEA.
//Portions created by Alex Shovkoplyas are
//Copyright (C) 2008 Alex Shovkoplyas. All Rights Reserved.
//------------------------------------------------------------------------------
unit BorPrj;

interface

uses
  SysUtils, Classes, IniFiles, JclBorlandTools, JclFileUtils;


type
  PUnitRecord = ^TUnitRecord;
  TUnitRecord = record Name, Path: TFileName; end;


  TBorlandProject = class
  protected
    FProjectFile: TFileName;
    FProjectDir: TFileName;
    FParsed: boolean;
    FIsLibrary: boolean;
    FUnits: array of TUnitRecord;
    FSearchPath: TFileName;
    FBinFile: TFileName;

    function GetUsesCount: integer;
    function GetUsesUnits(Index: integer): PUnitRecord;
    procedure ParseOptionsFile;
    procedure ParseProjectFile;
    procedure FixUnitPaths;
  public
    constructor Create(AProjectFile: TFileName);

    property ProjectFile: TFileName read FProjectFile;
    property Parsed: boolean read FParsed;
    property IsLibrary: boolean read FIsLibrary;
    property UsesCount: integer read GetUsesCount;
    property UsesUnits[Index: integer]: PUnitRecord read GetUsesUnits;
    property SearchPath: TFileName read FSearchPath;
    property BinFile: TFileName read FBinFile;
  end;


function BuildAbsPath(RelPath, BaseDir: TFileName): TFileName;


implementation


//------------------------------------------------------------------------------
//                               system
//------------------------------------------------------------------------------

{ TBorlandProject }

constructor TBorlandProject.Create(AProjectFile: TFileName);
begin
  FProjectFile := AProjectFile;
  if not FileExists(AProjectFile) then Exit;

  FProjectDir := ExtractFileDir(FProjectFile);

  try ParseProjectFile; except end;
  FixUnitPaths;
  ParseOptionsFile;
end;


function TBorlandProject.GetUsesCount: integer;
begin
  Result := Length(FUnits);
end;


function TBorlandProject.GetUsesUnits(Index: integer): PUnitRecord;
begin
  if (Index < 0) or (Index >= UsesCount)
    then Result := nil
    else Result := @FUnits[Index];
end;


procedure TBorlandProject.ParseOptionsFile;
const
  DelphiOptionsFileExtension = '.dof';
  DOFOutputDirName = 'OutputDir';
var
  OptionsFile: TFileName;
  OutBinDir: TFileName;
begin
  FSearchPath := '';
  FBinFile := ChangeFileExt(FProjectFile, BinaryExtensionExecutable);

  OptionsFile := ChangeFileExt(FProjectFile, DelphiOptionsFileExtension);
  if not FileExists(OptionsFile) then Exit;

  try
    with TIniFile.Create(OptionsFile) do
      try
        FSearchPath := ReadString(DOFDirectoriesSection, DOFSearchPathName, '');
        OutBinDir := ReadString(DOFDirectoriesSection, DOFOutputDirName, '');
      finally Free; end;
  except end;

  if OutBinDir <> '' then
    begin
    OutBinDir := BuildAbsPath(OutBinDir, FProjectDir);
    FBinFile := PathAppend(OutBinDir, ExtractFileName(FBinFile));
    end;
end;


function BuildAbsPath(RelPath, BaseDir: TFileName): TFileName;
begin
  Result := RelPath;
  if (Result = '') or PathIsAbsolute(Result) then Exit;

  BaseDir := ExcludeTrailingBackslash(BaseDir);

  while Copy(RelPath, 1, 3) = '..\' do
    begin
    BaseDir := ExtractFileDir(BaseDir);
    Delete(RelPath, 1, 3);
    end;

  Result := PathAppend(BaseDir, RelPath);
end;


procedure TBorlandProject.FixUnitPaths;
var
  i: integer;
begin
  for i:=0 to High(FUnits) do
    FUnits[i].Path := BuildAbsPath(FUnits[i].Path, FProjectDir);
end;






//------------------------------------------------------------------------------
//               parsing helper functions from JclParseUnits
//------------------------------------------------------------------------------
const
  Blanks: TSysCharSet = [#9, #10, #13, ' '];
  SLibrary = 'library';
  SProgram = 'program';
  SUnit = 'unit';
  SUses = 'uses';


function CheckKeyword(var P: PChar; Keyword: PChar): Boolean;
var
  KeywordLen: Integer;
begin
  KeywordLen := StrLen(Keyword);
  Result := StrLComp(P, Keyword, KeywordLen) = 0;
  if Result then Inc(P, KeywordLen);
end;


function CheckIdentifier(var P: PChar): Boolean;
begin
  Result := P^ in ['A'..'Z', '_', 'a'..'z'];
  if not Result then Exit;

  Inc(P);
  while P^ in ['0'..'9', 'A'..'Z', '_', 'a'..'z'] do Inc(P);
end;


function PeekKeyword(var P: PChar; Keyword: PChar): Boolean;
var
  KeywordLen: Integer;
begin
  KeywordLen := StrLen(Keyword);
  Result := StrLComp(P, Keyword, KeywordLen) = 0;
end;


procedure SkipChars(var P: PChar; Chars: TSysCharSet);
begin
  while P^ in Chars do Inc(P);
end;


procedure SkipComments(var P: PChar);
var
  Test: PChar;
begin
  if P^ = '{' then
    begin
    Test := StrScan(P, '}');
    if Test <> nil then P := Test + 1;
    end

  else if StrLComp(P, '(*', 2) = 0 then
    begin
    Test := StrPos(P, '*)');
    if Test <> nil then P := Test + 2;
    end

  else if StrLComp(P, '//', 2) = 0 then
    begin
    Test := StrPos(P, #13#10);
    if Test <> nil then P := Test + 2;
    end;
end;


procedure SkipCommentsAndBlanks(var P: PChar);
var
  Test: PChar;
begin
  repeat
    Test := P;
    SkipChars(P, Blanks);
    SkipComments(P);
  until Test = P;
end;






//------------------------------------------------------------------------------
//                          parse project file
//------------------------------------------------------------------------------
procedure TBorlandProject.ParseProjectFile;
var
  Text: string;
  P, PStart: PChar;
begin
  //read project file
  with TFileStream.Create(FProjectFile, fmOpenRead or fmShareDenyWrite) do
    try
      SetLength(Text, Size);
      if Size > 0 then ReadBuffer(Text[1], Size);
    finally Free; end;

  P := PChar(Text);

  //'program' or 'library'
  SkipCommentsAndBlanks(P);
  FIsLibrary := PeekKeyword(P, SLibrary);
  if not (FIsLibrary or PeekKeyword(P, SProgram)) then Abort;
  if not CheckIdentifier(P) then Abort;

  //name
  SkipCommentsAndBlanks(P);
  if not CheckIdentifier(P) then Abort;

  // ;
  SkipCommentsAndBlanks(P);
  if P^ <> ';' then Abort;
  Inc(P);

  //'uses'
  SkipCommentsAndBlanks(P);
  if not CheckKeyword(P, SUses) then
    begin FParsed := true; Exit; end;

  //unit hames
  while P^ <> #0 do
    begin
    SkipCommentsAndBlanks(P);

    PStart := P;
    if not CheckIdentifier(P) then Abort;
    SetLength(FUnits, Length(FUnits)+1);
    SetString(FUnits[High(FUnits)].Name, PStart, P - PStart);

    SkipCommentsAndBlanks(P);

    if PeekKeyword(P, 'in') then
      begin
      Inc(P, 2);
      SkipCommentsAndBlanks(P);
      if P^ <> '''' then Abort;
      Inc(P);
      PStart := P;

      while not (P^ in [#0, '''']) do Inc(P);
      if P^ <> '''' then Abort;
      SetString(FUnits[High(FUnits)].Path, PStart, P - PStart);
      Inc(P);
      SkipCommentsAndBlanks(P);
      end;

      case P^ of
        ',': Inc(P);
        ';': begin Inc(P); Break; end;
        else Abort;
        end;
    end;  

  FParsed := true;
end;



end.

