//------------------------------------------------------------------------------
//The contents of this file are subject to the Mozilla Public License
//Version 1.1 (the "License"); you may not use this file except in compliance
//with the License. You may obtain a copy of the License at
//http://www.mozilla.org/MPL/ Software distributed under the License is
//distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
//or implied. See the License for the specific language governing rights and
//limitations under the License.
//
//The Original Code is DbgSrc.pas.
//
//The Initial Developer of the Original Code is Alex Shovkoplyas, VE3NEA.
//Portions created by Alex Shovkoplyas are
//Copyright (C) 2008 Alex Shovkoplyas. All Rights Reserved.
//------------------------------------------------------------------------------
unit DbgSrc;

interface

uses
  Windows, SysUtils, DbgHelp, JclDebug, PathUtil;


type
  TDebugInfoKind = (diNone, diExpTable, diMsSymb, diBorSymb);


const
  DbgSInfoKindName: array[TDebugInfoKind] of string =
    ('None', 'Export Table', 'MS Symbols', 'MAP file');

  STR_UNKNOWN = '<Unknown>';


type
  TDebugInfoSource = class(TJclDebugInfoSource)
  protected
    FKind: TDebugInfoKind;
    procedure ClearInfo(var Info: TJclLocationInfo);
  public
    BinFileName: TFileName;

    function InitializeSource: Boolean; override;
    function GetLocationInfo(const Addr: Pointer; var Info: TJclLocationInfo): Boolean; override;
    function GetFuncAddress(AFuncName: string): DWord; virtual;
    function GetNextFunc(Addr: DWord; var Info: TJclLocationInfo): Boolean; virtual;

    property Kind: TDebugInfoKind read FKind;
  end;


  TDebugInfoMap = class(TDebugInfoSource)
  protected
    FScanner: TJclMapScanner;
  public
    destructor Destroy; override;
    function InitializeSource: Boolean; override;
    function GetLocationInfo(const Addr: Pointer; var Info: TJclLocationInfo): Boolean; override;
  end;


  TDebugInfoSymbols = class(TDebugInfoSource)
  protected
    ModuleInfo: TImagehlpModule;
    Sp: TImageHlpSymbolPackage;
  public
    function InitializeSource: Boolean; override;
    function GetLocationInfo(const Addr: Pointer; var Info: TJclLocationInfo): Boolean; override;
    function GetFuncAddress(AFuncName: string): DWord; override;
    function GetNextFunc(Addr: DWord; var Info: TJclLocationInfo): Boolean; override;
  end;


function CreateDebugInfoSource(ABinFile: TFileName; ABaseAddr: DWord): TDebugInfoSource;
procedure SetSymbolsPath(APath: TFileName; ACanDownload: boolean);


var
  DummyDebugInfoSource: TDebugInfoSource;



implementation

var
  SymbolsProcHandle: THandle;
  SymbolsAvailable: boolean;


//------------------------------------------------------------------------------
//                  find the best debug source for the file
//------------------------------------------------------------------------------
function CreateDebugInfoSource(ABinFile: TFileName; ABaseAddr: DWord): TDebugInfoSource;
begin
  //try MAP
  Result := TDebugInfoMap.Create(ABaseAddr);
  with TDebugInfoMap(Result) do
    begin
    BinFileName := ABinFile;
    if not InitializeSource then FreeAndNil(Result);
    end;
  if Result <> nil then Exit;


  //try PDB symbols
  Result := TDebugInfoSymbols.Create(ABaseAddr);
  with TDebugInfoSymbols(Result) do
    begin
    BinFileName := ABinFile;
    if not InitializeSource then FreeAndNil(Result);
    end;
  if Result <> nil then Exit;

  //no debug info found
  Result := DummyDebugInfoSource;
end;






//------------------------------------------------------------------------------
//                            dummy source
//------------------------------------------------------------------------------

{ TDebugInfo }

function TDebugInfoSource.InitializeSource: Boolean;
begin
  Result := true;
end;


procedure TDebugInfoSource.ClearInfo(var Info: TJclLocationInfo);
begin
  Info.Address := nil;
  Info.UnitName := STR_UNKNOWN;
  Info.ProcedureName := STR_UNKNOWN;
  Info.OffsetFromProcName := 0;
  Info.LineNumber := 0;
  Info.OffsetFromLineNumber := 0;
end;


function TDebugInfoSource.GetLocationInfo(const Addr: Pointer; var Info: TJclLocationInfo): Boolean;
begin
  ClearInfo(Info);
  Info.Address := Addr;
  Result := false;
end;


function TDebugInfoSource.GetFuncAddress(AFuncName: string): DWord;
begin
  Result := 0;
end;


function TDebugInfoSource.GetNextFunc(Addr: DWord; var Info: TJclLocationInfo): Boolean;
begin
  ClearInfo(Info);
  Result := false;
end;






//------------------------------------------------------------------------------
//                           Delphi MAP file
//------------------------------------------------------------------------------

{ TDebugInfoMap }

destructor TDebugInfoMap.Destroy;
begin
  FScanner.Free;
  inherited;
end;


function TDebugInfoMap.InitializeSource: Boolean;
var
  MapFileName: TFileName;
begin
  MapFileName := ChangeFileExt(BinFileName, JclMapFileExtension);
  Result := FileExists(MapFileName);
  if not Result then Exit;

  FScanner := TJclMapScanner.Create(MapFileName, Module);
  FKind := diBorSymb;
end;


function TDebugInfoMap.GetLocationInfo(const Addr: Pointer; var Info: TJclLocationInfo): Boolean;
var
  VA: DWORD;
begin
  ClearInfo(Info);
  Info.Address := Addr;

  VA := VAFromAddr(Addr);

  with FScanner do
    begin
    Info.UnitName := ModuleNameFromAddr(VA);

    Result := Info.UnitName <> '';
    if Result
      then
        begin
        Info.ProcedureName := ProcNameFromAddr(VA, Info.OffsetFromProcName);
        Info.LineNumber := LineNumberFromAddr(VA, Info.OffsetFromLineNumber);
        Info.SourceName := SourceNameFromAddr(VA);
        end
      else
        Info.UnitName := STR_UNKNOWN;
    end;
end;






//------------------------------------------------------------------------------
//                              MS symbols
//------------------------------------------------------------------------------
procedure SetSymbolsPath(APath: TFileName; ACanDownload: boolean);
const
  ServerPath = 'http://msdl.microsoft.com/download/symbols';
begin
  if ACanDownload then APath := APath + ';SRV*' + APath + '*' +ServerPath;
  SymSetSearchPath(SymbolsProcHandle, PChar(APath));
end;


function InitializeDebugSymbols: boolean;
const
  GOOD_OPTIONS =
   SYMOPT_CASE_INSENSITIVE or
   SYMOPT_UNDNAME or
   SYMOPT_LOAD_LINES or
   SYMOPT_OMAP_FIND_NEAREST or
   //SYMOPT_LOAD_ANYTHING or
   SYMOPT_NO_UNQUALIFIED_LOADS or
   SYMOPT_FAIL_CRITICAL_ERRORS or
   SYMOPT_INCLUDE_32BIT_MODULES or
   //SYMOPT_DEBUG or
   SYMOPT_OVERWRITE;
begin
  //unique non-zero value
  SymbolsProcHandle := GetCurrentThreadId + $555555;

  SymbolsAvailable := SymInitialize(SymbolsProcHandle, nil, false);
  Result := SymbolsAvailable;
  if not Result then Exit;

  SymSetOptions(GOOD_OPTIONS);
end;


procedure CleanupDebugSymbols;
begin
  if SymbolsAvailable then SymCleanup(SymbolsProcHandle);
end;


function TDebugInfoSymbols.InitializeSource: Boolean;
var
  rc: DWord;
begin
  Result := SymbolsAvailable;
  if not Result then Exit;

  //unload old data
  if FKind <> diMsSymb then SymUnloadModule(SymbolsProcHandle, Module);

  //load module
  rc := SymLoadModule(SymbolsProcHandle, 0, PChar(BinFileName), nil, Module, 0);
  Result := (rc <> 0) or (GetLastError = ERROR_SUCCESS);
  if not Result then Exit;

  //get module info
  ZeroMemory(@ModuleInfo, SizeOf(ModuleInfo));
  ModuleInfo.SizeOfStruct := SizeOf(TImagehlpModule);
  Result := SymGetModuleInfo(SymbolsProcHandle, Module, ModuleInfo);

  //info type
  case ModuleInfo.SymType of
    SymCoff, SymCv, SymPdb, SymSym, SymDia: FKind := diMsSymb;
    SymExport: FKind := diExpTable;
    end;
  Result := Result and (FKind <> diNone);

  if not Result then Exit;
end;


function TDebugInfoSymbols.GetLocationInfo(const Addr: Pointer;
  var Info: TJclLocationInfo): Boolean;
var
  Displacement: integer;
  Line: TImageHelpLine;
begin
  ClearInfo(Info);
  Info.Address := Addr;

  Result := SymbolsAvailable;
  if not Result then Exit;

  ZeroMemory(@Sp, SizeOf(Sp));
  Sp.sym.SizeOfStruct := SizeOf(Sp.sym);
  Sp.sym.MaxNameLength := MAX_SYM_NAME;
  Displacement := 0;
  Result := SymGetSymFromAddr(SymbolsProcHandle, DWORD(Addr), @Displacement, Sp.sym);
  if not Result then Exit;

  Info.ProcedureName := PChar(@Sp.sym.Name);
  Info.OffsetFromProcName := Displacement;

  ZeroMemory(@Line, SizeOf(Line));
  Line.SizeOfStruct := SizeOf(Line);
  Displacement := 0;
  if not SymGetLineFromAddr(SymbolsProcHandle, DWORD(Addr), @Displacement, Line) then Exit;

  Info.UnitName := Line.FileName;
  Info.LineNumber := Line.LineNumber;
  Info.OffsetFromLineNumber := Displacement;
end;


function TDebugInfoSymbols.GetFuncAddress(AFuncName: string): DWord;
begin
  Result := 0;

  AFuncName := ModuleInfo.ModuleName + '!' + AFuncName;

  ZeroMemory(@Sp, SizeOf(Sp));
  Sp.Sym.SizeOfStruct := SizeOf(Sp.Sym);
  Sp.Sym.MaxNameLength := MAX_SYM_NAME;

  if not SymGetSymFromName(SymbolsProcHandle, PChar(AFuncName), Sp.Sym) then Exit;

  Result := Sp.Sym.Address;
end;


function TDebugInfoSymbols.GetNextFunc(Addr: DWord; var Info: TJclLocationInfo): Boolean;
begin
  ClearInfo(Info);

  Result := SymGetSymNext(SymbolsProcHandle, Sp.Sym);

  if Result then
    begin
    Info.Address := Ptr(Sp.Sym.Address);
    Info.ProcedureName := PChar(@Sp.sym.Name);
    end;
end;






//------------------------------------------------------------------------------
//                       instantiate dummy source
//------------------------------------------------------------------------------

initialization
  DummyDebugInfoSource := TDebugInfoSource.Create(0);
  InitializeDebugSymbols;

finalization
  DummyDebugInfoSource.Free;
  CleanupDebugSymbols;

end.

