//------------------------------------------------------------------------------
//The contents of this file are subject to the Mozilla Public License
//Version 1.1 (the "License"); you may not use this file except in compliance
//with the License. You may obtain a copy of the License at
//http://www.mozilla.org/MPL/ Software distributed under the License is
//distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
//or implied. See the License for the specific language governing rights and
//limitations under the License.
//
//The Original Code is FileFind.pas.
//
//The Initial Developer of the Original Code is Alex Shovkoplyas, VE3NEA.
//Portions created by Alex Shovkoplyas are
//Copyright (C) 2008 Alex Shovkoplyas. All Rights Reserved.
//------------------------------------------------------------------------------
unit FileFind;

interface

uses
  SysUtils, Classes, JclBorlandTools, JclFileUtils, BorPrj, Math;


const
  STR_NOTFOUND = '<Not Found>';
    

type
  TSourceFileFinder = class
  private
    Tools: TJclBorRADToolInstallations;
    SearchDirs: TStringList;
    FBinFile: TFileName;
    FBorToolIndex: integer;
    Prj: TBorlandProject;
    procedure Clear;
    procedure SetBorToolIndex(const Value: integer);
    function IsSameProject(AProjectFile: TFileName): boolean;
    function GetProjFile: TFileName;
    procedure ListSearchDirs;
    function IsSameBinFile(ABinFile: TFileName): boolean;
  public
    constructor Create;
    destructor Destroy; override;
    function GetToolNames: string;
    procedure SetProject(AFileName: TFileName);
    function GetPath(AUnit: string): TFileName;

    property BorToolIndex: integer read FBorToolIndex write SetBorToolIndex;
    property BinFile: TFileName read FBinFile;
    property ProjFile: TFileName read GetProjFile;
  end;





implementation

//------------------------------------------------------------------------------
//                           helper functions
//------------------------------------------------------------------------------
function ProjectNameFromBinName(ABinFile: TFileName): TFileName;
const
  ProjExt: array[0..3] of string = (
    SourceExtensionDelphiProject, SourceExtensionBCBProject,
    SourceExtensionBDSProject, SourceExtensionDProject);
var
  i: integer;
begin
  for i:=0 to High(ProjExt) do
    begin
    Result := ChangeFileExt(ABinFile, ProjExt[i]);
    if FileExists(Result) then Exit;
    end;
  Result := '';
end;






//------------------------------------------------------------------------------
//                                system
//------------------------------------------------------------------------------

{ TSourceFileFinder }

constructor TSourceFileFinder.Create;
begin
  Tools := TJclBorRADToolInstallations.Create;
  SearchDirs := TStringList.Create;

  FBorToolIndex := Tools.Count-1;
end;


destructor TSourceFileFinder.Destroy;
begin
  Tools.Free;
  SearchDirs.Free;
  Prj.Free;
  inherited;
end;


procedure TSourceFileFinder.Clear;
begin
  FBinFile := '';
  FreeAndNil(Prj);
end;


function TSourceFileFinder.GetToolNames: string;
var
  i: integer;
begin
  Result := '';
  for i:=0 to Tools.Count-1 do
    Result := Result + Tools.Installations[i].Name + #13;

  Result := Trim(Result);
end;






//------------------------------------------------------------------------------
//                                get/set
//------------------------------------------------------------------------------
function TSourceFileFinder.GetProjFile: TFileName;
begin
  if Prj = nil then Result := '' else Result := Prj.ProjectFile;
end;


function TSourceFileFinder.IsSameProject(AProjectFile: TFileName): boolean;
begin
  Result := CompareText(ProjFile, AProjectFile) = 0;
end;


function TSourceFileFinder.IsSameBinFile(ABinFile: TFileName): boolean;
begin
  Result := CompareText(FBinFile, ABinFile) = 0;
end;


procedure TSourceFileFinder.SetBorToolIndex(const Value: integer);
begin
  if Value = FBorToolIndex then Exit;
  FBorToolIndex := Max(0, Min(Tools.Count-1, Value));
  Clear;
end;


function IsFileBinary(AFileName: TFileName): boolean;
const
  BinaryExtensions: array[0..2] of string = (
    BinaryExtensionExecutable, BinaryExtensionLibrary, '.ocx');
var
  Ext: string;
  i: integer;
begin
  Result := true;
  Ext := LowerCase(ExtractFileExt(AFileName));
  for i:=0 to High(BinaryExtensions) do
    if Ext = BinaryExtensions[i] then Exit;
  Result := false;
end;


procedure TSourceFileFinder.SetProject(AFileName: TFileName);
var
  NewProjFile: TFileName;
begin
  if IsFileBinary(AFileName)
    then
      //if dll/exe, try to find dpr; if found dpr, create Prj; preserve binfile
      begin
      if IsSameBinFile(AFileName) then Exit;
      NewProjFile := ProjectNameFromBinName(AFileName);
      SetProject(NewProjFile);
      FBinFile := AFileName;
      if Prj = nil then ListSearchDirs;
      end
    else
      //if dpr, create Prj and extract binfile
      begin
      if IsSameProject(AFileName) then Exit;
      Clear;
      Prj := TBorlandProject.Create(AFileName);
      if Prj.Parsed
        then FBinFile := Prj.BinFile
        else begin FBinFile := ''; FreeAndNil(Prj); end;

      ListSearchDirs;
      end;
end;


procedure TSourceFileFinder.ListSearchDirs;
var
  S: string;
begin
  SearchDirs.Clear;

  //project folder
  if Prj <> nil
    then S := ExtractFileDir(Prj.ProjectFile)
    else S := ExtractFileDir(FBinFile);

  //search path in project settings
  if Prj <> nil then
    PathListIncludeItems(S, Prj.SearchPath);

  //Delphi lib and browsing path
  if FBorToolIndex >= 0 then
    begin
    PathListIncludeItems(S, Tools.Installations[FBorToolIndex].LibrarySearchPath);
    PathListIncludeItems(S, Tools.Installations[FBorToolIndex].LibraryBrowsingPath);
    end;

  //expand vars like $(DELPHI)
  S := Tools.Installations[FBorToolIndex].SubstitutePath(S);

  SearchDirs.Text := StringReplace(S, ';', #13, [rfReplaceAll]);
end;


function TSourceFileFinder.GetPath(AUnit: string): TFileName;
var
  FileName: TFileName;
  i: integer;
begin
  //return project itself
  if CompareText(AUnit, ChangeFileExt(ExtractFileName(ProjFile), '')) = 0 then
    begin Result := ProjFile; Exit; end;


  //use explicit path if specified in .dpr
  Result := '';

  if Prj <> nil then
    for i:=0 to Prj.UsesCount-1 do
      if CompareText(Prj.UsesUnits[i].Name, AUnit) = 0 then
        begin
        Result := Prj.UsesUnits[i].Path;
        Break;
        end;

  if Result <> '' then Exit;


  //try all search dirs        
  FileName := ChangeFileExt(AUnit, SourceExtensionPAS);
  for i:=0 to SearchDirs.Count-1 do
    begin
    Result := PathAppend(SearchDirs[i], FileName);
    if FileExists(Result) then Exit;
    end;

  FileName := ChangeFileExt(AUnit, SourceExtensionCPP);
  for i:=0 to SearchDirs.Count-1 do
    begin
    Result := PathAppend(SearchDirs[i], FileName);
    if FileExists(Result) then Exit;
    end;

  Result := STR_NOTFOUND;
end;



end.

