//------------------------------------------------------------------------------
//The contents of this file are subject to the Mozilla Public License
//Version 1.1 (the "License"); you may not use this file except in compliance
//with the License. You may obtain a copy of the License at
//http://www.mozilla.org/MPL/ Software distributed under the License is
//distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
//or implied. See the License for the specific language governing rights and
//limitations under the License.
//
//The Original Code is HitCnts.pas.
//
//The Initial Developer of the Original Code is Alex Shovkoplyas, VE3NEA.
//Portions created by Alex Shovkoplyas are
//Copyright (C) 2008 Alex Shovkoplyas. All Rights Reserved.
//------------------------------------------------------------------------------
unit HitCnts;

interface

uses
  Windows, SysUtils, Classes, KeyedLst, JclDebug, ModulInf;


type
  TCustomHitCounter = class;

  
  PSample = ^TSample;
  TSample = record
    Addr: DWord;
    Thr: DWord;
    case boolean of
      false: (Cpu: DWord; Mode: DWord;);
      true:  (Count: integer; HitCounter: TCustomHitCounter);
    end;
  TSamples = array of TSample;


  TCustomHitCounter = class
  protected
    FName: string;
    FChildren: TStringList;
    function GetName: string; virtual;
    function GetChildCount: integer; virtual;
    function GetChild(Index: integer): TCustomHitCounter; virtual;
  public
    Hits: integer;
    ThrHits: array of Integer;
    Parent: TCustomHitCounter;
    FirstAddr: DWord;

    constructor Create(AParent: TCustomHitCounter);
    destructor Destroy; override;
    procedure AddHits(const Sample: TSample; const Location: TJclLocationInfo); virtual;

    property Name: string read GetName write FName;
    property ChildCount: integer read GetChildCount;
    property Children[Index: integer]: TCustomHitCounter read GetChild;
  end;


  TAddrHits = class(TCustomHitCounter)
  public
    Addr: DWord;
  end;


  TLineHits = class(TCustomHitCounter)
  public
    LineNo: integer;
    function GetName: string; override;
    procedure AddHits(const Sample: TSample; const Location: TJclLocationInfo); override;
  end;


  TFuncHits = class(TCustomHitCounter)
  public
    procedure AddHits(const Sample: TSample; const Location: TJclLocationInfo); override;
  end;


  TUnitHits = class(TCustomHitCounter)
  public
    SrcFile: TFileName;
    procedure AddHits(const Sample: TSample; const Location: TJclLocationInfo); override;
  end;

  TModuleHits = class(TCustomHitCounter)
  public
    ModuleInfo: TModuleItem;
    Mode: DWord;
    constructor Create(ThrCount: integer);
    procedure AddHits(const Sample: TSample; const Location: TJclLocationInfo); override;
  end;




implementation

{ TCustomHitCounter }

constructor TCustomHitCounter.Create(AParent: TCustomHitCounter);
begin
  FChildren := TStringList.Create;
  FChildren.Sorted := true;
  Parent := AParent;
  FirstAddr := $FFFFFFFF;

  if Parent <> nil then SetLength(ThrHits, Length(Parent.ThrHits));
end;


destructor TCustomHitCounter.Destroy;
var
  i: integer;
begin
  for i:=0 to FChildren.Count-1 do FChildren.Objects[i].Free;
  FChildren.Free;
  inherited;
end;


function TCustomHitCounter.GetName: string;
begin
  Result := FName;
end;


function TCustomHitCounter.GetChildCount: integer;
begin
  Result := FChildren.Count;
end;


function TCustomHitCounter.GetChild(Index: integer): TCustomHitCounter;
begin
  Result := TCustomHitCounter(FChildren.Objects[Index]);
end;


procedure TCustomHitCounter.AddHits(const Sample: TSample;
  const Location: TJclLocationInfo);
begin
  Inc(Hits, Sample.Count);
  Inc(ThrHits[Sample.Thr], Sample.Count);
  if Sample.Addr < FirstAddr then FirstAddr := Sample.Addr;
end;


{ TLineHits }

procedure TLineHits.AddHits(const Sample: TSample;
  const Location: TJclLocationInfo);
var
  AddrName: string;
  AddrHits: TAddrHits;
  Idx: integer;
begin
  inherited;

  AddrName := Format('%.8x', [Sample.Addr]);

  if FChildren.Find(AddrName, Idx)
    then
      AddrHits := TAddrHits(FChildren.Objects[Idx])
    else
      begin
      AddrHits := TAddrHits.Create(Self);
      AddrHits.Addr := Sample.Addr;
      AddrHits.Name := AddrName;
      FChildren.AddObject(AddrName, AddrHits);
      end;   

  AddrHits.AddHits(Sample, Location);
end;


function TLineHits.GetName: string;
begin
  if LineNo = 0
    then Result := STR_UNKNOWN
    else Result := Format('Line %d', [LineNo]);
end;



{ TFuncHits }

procedure TFuncHits.AddHits(const Sample: TSample; const Location: TJclLocationInfo);
var
  LineName: string;
  Idx: integer;
  LineHits: TLineHits;
begin
  inherited;

  LineName := Format('%.8x', [Location.LineNumber]);

  if FChildren.Find(LineName, Idx)
    then
      LineHits := TLineHits(FChildren.Objects[Idx])
    else
      begin
      LineHits := TLineHits.Create(Self);
      LineHits.LineNo := Location.LineNumber;
      LineHits.Name := LineName;
      FChildren.AddObject(LineName, LineHits);
      end;

  LineHits.AddHits(Sample, Location);
end;



{ TUnitHits }

procedure TUnitHits.AddHits(const Sample: TSample; const Location: TJclLocationInfo);
var
  FuncName: string;
  Idx: integer;
  FuncHits: TFuncHits;
begin
  inherited;

  FuncName := Location.ProcedureName;
  if FuncName = '' then FuncName := STR_UNKNOWN;

  if FChildren.Find(FuncName, Idx)
    then
      FuncHits := TFuncHits(FChildren.Objects[Idx])
    else
      begin
      FuncHits := TFuncHits.Create(Self);
      FuncHits.Name := FuncName;
      FChildren.AddObject(FuncName, FuncHits);
      end;

  FuncHits.AddHits(Sample, Location);
end;



{ TModuleHits }

constructor TModuleHits.Create(ThrCount: integer);
begin
  inherited Create(nil);
  SetLength(ThrHits, ThrCount);
end;


procedure TModuleHits.AddHits(const Sample: TSample; const Location: TJclLocationInfo);
var
  UnitName: string;
  Idx: integer;
  UnitHits: TUnitHits;
begin
  inherited;

  Mode := Mode or Sample.Mode;

  UnitName := Location.UnitName;
  if UnitName = '' then UnitName := STR_UNKNOWN;

  if FChildren.Find(UnitName, Idx)
    then
      UnitHits := TUnitHits(FChildren.Objects[Idx])
    else
      begin
      UnitHits := TUnitHits.Create(Self);
      UnitHits.Name := UnitName;
      FChildren.AddObject(UnitName, UnitHits);
      end;

  UnitHits.AddHits(Sample, Location);
end;



end.

