//------------------------------------------------------------------------------
//The contents of this file are subject to the Mozilla Public License
//Version 1.1 (the "License"); you may not use this file except in compliance
//with the License. You may obtain a copy of the License at
//http://www.mozilla.org/MPL/ Software distributed under the License is
//distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
//or implied. See the License for the specific language governing rights and
//limitations under the License.
//
//The Original Code is ModulInf.pas.
//
//The Initial Developer of the Original Code is Alex Shovkoplyas, VE3NEA.
//Portions created by Alex Shovkoplyas are
//Copyright (C) 2008 Alex Shovkoplyas. All Rights Reserved.
//------------------------------------------------------------------------------
unit ModulInf;

interface

uses
  Windows, SysUtils, Classes, PsApi, KeyedLst, JclDebug, DbgSrc, PathUtil;


const
  STR_UNKNOWN = '<Unknown>';


type
  TCodeSection = record
    Name: string;
    StartAddr, EndAddr: DWord;
    end;


  TModuleItem = class(TKeyedItem)
  private
    FDebugInfoSource: TDebugInfoSource;
    function GetDebugInfoSource: TDebugInfoSource;
  public
    Main: boolean;
    Name: TFileName;
    BinFile: TFileName;
    BinExists: boolean;
    EndAddr: DWord;

    destructor Destroy; override;
    property BaseAddr: DWord read Key write Key;
    function IsAddressInModule(AAddr: DWord): boolean;

    property DebugInfoSource: TDebugInfoSource read GetDebugInfoSource;
  end;


  TModuleList = class(TKeyedList)
  private
    procedure GetAddresses(Module: TModuleItem);
    function GetItem(Index: integer): TModuleItem;
  public
    procedure ListDrivers;
    procedure AddModule(BaseAddr: Pointer; FileHandle: THandle; Main: boolean = false);
    procedure GetMissingData;
    function GetModuleIndexFromAddr(AAddr: DWord): integer;

    property Items[Index: integer]: TModuleItem read GetItem; default;
    end;



implementation


{ TModuleItem }

function TModuleItem.GetDebugInfoSource: TDebugInfoSource;
begin
  if FDebugInfoSource = nil then
    FDebugInfoSource := CreateDebugInfoSource(BinFile, BaseAddr);
  Result := FDebugInfoSource;
end;


function TModuleItem.IsAddressInModule(AAddr: DWord): boolean;
begin
  Result := (AAddr >= BaseAddr) and (AAddr < EndAddr);
end;


destructor TModuleItem.Destroy;
begin
  if FDebugInfoSource <> DummyDebugInfoSource then FDebugInfoSource.Free;
  inherited;
end;





{ TModuleList }

//------------------------------------------------------------------------------
//                             get module list
//------------------------------------------------------------------------------
procedure TModuleList.ListDrivers;
var
  Arr: array[0..2048] of Pointer;
  Cnt: DWord;
  i: integer;
  S: TFileName;
  Module: TModuleItem;
begin
  Clear;

  //list drivers
  if not EnumDeviceDrivers(@Arr, 2048, Cnt) then Exit;
  Cnt := Cnt div SizeOf(Pointer);

  //get driver info
  for i:=0 to Cnt-1 do
    begin
    Module := TModuleItem.Create;
    Module.BaseAddr := DWord(Arr[i]);

    SetLength(S, MAX_PATH);
    SetLength(S, GetDeviceDriverFileName(Arr[i], @S[1], MAX_PATH));
    Module.BinFile := Trim(S);

    Add(Module);
    end;
end;


procedure TModuleList.AddModule(BaseAddr: Pointer; FileHandle: THandle; Main: boolean = false);
var
  Module: TModuleItem;
begin
    Module := TModuleItem.Create;
    Module.BaseAddr := DWord(BaseAddr);
    Module.BinFile := FileNameFromHandle(FileHandle);

    Add(Module);
end;


function TModuleList.GetItem(Index: integer): TModuleItem;
begin
  Result := TModuleItem(inherited Items[Index]);
end;






//------------------------------------------------------------------------------
//                           get module info
//------------------------------------------------------------------------------
procedure TModuleList.GetMissingData;
var
  i: integer;
  Module: TModuleItem;
begin
  //paths and names
  for i:=0 to Count-1 do
    with Items[i] do
     begin
     Name := ExtractFileName(BinFile);
     BinFile := FixPath(BinFile);
     BinExists := FileExists(BinFile);
     end;

  //addresses
  for i:=0 to Count-1 do
    with Items[i] do
     if BinExists then GetAddresses(Items[i]);

  //missing addresses
  for i:=0 to Count-1 do
    with Items[i] do
      begin
      if EndAddr = 0 then
        if i < (Count-1)
          then EndAddr := Items[i+1].BaseAddr
          else EndAddr := $FFFFFFFF;
      end;

  //dummy module for all unknown addresses
  Module := TModuleItem.Create;
  Module.Name := STR_UNKNOWN;
  Add(Module);
end;


procedure TModuleList.GetAddresses(Module: TModuleItem);
const
  IMAGE_NT_OPTIONAL_HDR32_MAGIC = $10b;
var
  Fs: TFileStream;
  DosHeader: TImageDosHeader;
  //PeHeader: TImageFileHeader;
  OptHeader: TImageOptionalHeader;
  PeMagic: integer;
begin
  try
    //read headers
    Fs := TFileStream.Create(Module.BinFile, fmOpenRead	or fmShareDenyNone);
    try
      //dos header
      Fs.ReadBuffer(DosHeader, SizeOf(TImageDosHeader));
      if DosHeader.e_magic <> IMAGE_DOS_SIGNATURE then Exit;
      //PE header
      Fs.Position := DosHeader._lfanew;
      Fs.ReadBuffer(PeMagic, SizeOf(Integer));
      if PeMagic <> IMAGE_NT_SIGNATURE then Exit;
      Fs.Position := Fs.Position + SizeOf(TImageFileHeader);
      //optional header
      Fs.ReadBuffer(OptHeader, SizeOf(TImageOptionalHeader));
      if OptHeader.Magic <> IMAGE_NT_OPTIONAL_HDR32_MAGIC then Exit;
    finally
      Fs.Free;
    end;

    //get addresses
    Module.EndAddr := Module.BaseAddr + OptHeader.SizeOfImage;
  except
  end;
end;






//------------------------------------------------------------------------------
//                           module at address
//------------------------------------------------------------------------------
function TModuleList.GetModuleIndexFromAddr(AAddr: DWord): integer;
begin
  Result := FindKey(AAddr);

  if Result = 0 then Exit
  else if Result = Count then begin Result := 0; Exit; end;

  if AAddr < Items[Result].BaseAddr then Dec(Result);

  if not Items[Result].IsAddressInModule(AAddr)
    then Result := 0;
end;



end.

