//------------------------------------------------------------------------------
//The contents of this file are subject to the Mozilla Public License
//Version 1.1 (the "License"); you may not use this file except in compliance
//with the License. You may obtain a copy of the License at
//http://www.mozilla.org/MPL/ Software distributed under the License is
//distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
//or implied. See the License for the specific language governing rights and
//limitations under the License.
//
//The Original Code is PerfData.pas.
//
//The Initial Developer of the Original Code is Alex Shovkoplyas, VE3NEA.
//Portions created by Alex Shovkoplyas are
//Copyright (C) 2008 Alex Shovkoplyas. All Rights Reserved.
//------------------------------------------------------------------------------
unit PerfData;

interface

uses
  Windows, SysUtils, Classes, KeyedLst, ThrLst, ModulInf, PerfUtil, JclDebug,
  HitCnts;


type
  TProcessModuleEvent = procedure(Sender: TObject; ModuleName: string; Percent: Single) of object;


  TPerformanceData = class
  private
    FOnProcessModule: TProcessModuleEvent;

    procedure AnalyzeSamples;
    procedure ResetCounters;
    procedure LoadFromNrm2FileApi(AFileName: TFileName);
    procedure LoadFromNrm2FileDirect(AFileName: TFileName);
  public
    Pid: DWord;
    ProcessName: string;
    CpuCount: integer;
    BufferFull: boolean;

    Modules: TModuleList;
    Threads: TThreadList;
    ThreadNumbers: array of integer;

    Samples: TSamples;
    TotalHits, ProcessHits: integer;
    CpuHits, ModeHits, ThrHits: array of integer;
    CpuThrHits, ModeThrHits: array of array of integer;
    HitsByModule: array of TModuleHits;


    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    procedure LoadSamples(AFileName: TFileName);
    procedure LoadDebugInfo;
    function IsEmpty:  boolean;
    function FindSample(Addr: DWord): integer;

    property OnProcessModule: TProcessModuleEvent read FOnProcessModule write FOnProcessModule;
  end;


implementation

{ TPerformanceData }

constructor TPerformanceData.Create;
begin
  Modules := TModuleList.Create;
  Threads := TThreadList.Create;
end;


destructor TPerformanceData.Destroy;
begin
  ResetCounters;
  Modules.Free;
  Threads.Free;

  inherited;
end;


procedure TPerformanceData.Clear;
begin
  Samples := nil;
  TotalHits := 0;
  ProcessHits := 0;
  ResetCounters;
  Modules.Clear;
  Threads.Clear;
end;


procedure TPerformanceData.LoadSamples(AFileName: TFileName);
begin
  LoadFromNrm2FileDirect(AFileName);
  AnalyzeSamples;
end;


//slow, use LoadFromNrm2FileDirect instead
procedure TPerformanceData.LoadFromNrm2FileApi(AFileName: TFileName);
var
  rc: integer;
  Hook: PTraceHook;
  Cnt: integer;
  Fi: TNrmFileInfo;
begin
  TotalHits := 0;
  ProcessHits := 0;
  Samples := nil;

  rc := Nrm2OpenTraceFile(PChar(AFileName));
  if rc <> 0 then raise Exception.Create('Unable to open the NRM2 file');

  FillChar(Fi, SizeOf(fi), 0);
  rc := Nrm2GetFileInfo(@Fi);
  if rc <> 0 then raise Exception.Create('Unable to read the NRM2 file');

  CpuCount := Fi.num_cpus;
  Cnt := 0;
  SetLength(Samples, 1024);

  try
    repeat
      Hook := Nrm2GetNextHook;
      if Hook = nil then Break;
      with Hook^ do
        begin
        if major <> SFT_HOOK_TPROF_MAJOR then Continue;
        Inc(TotalHits);
        if ival[0] <> Pid then Continue;
        Inc(ProcessHits);

        Samples[Cnt].Thr := ival[1];
        Samples[Cnt].Addr := ival[2];
        Samples[Cnt].Cpu := cpu_no;
        Samples[Cnt].Mode := 1 - (minor and 1);

        Inc(Cnt);
        if Cnt = Length(Samples) then SetLength(Samples, Length(Samples) * 2);
        end;
    until false;
  finally
    Nrm2CloseTraceFile;
    SetLength(Samples, Cnt);
  end;
end;


procedure TPerformanceData.LoadFromNrm2FileDirect(AFileName: TFileName);
type
  PPerfHeader = ^TPerfHeader;
  TPerfHeader = record
    nrm2: DWord;
    Dummy1: array[0..1] of DWord;
    header_size: DWord;
    data_size: DWord;
    Dummy2: DWord;
    cpu_no: DWord;
    Dummy3: array[0..6] of DWord;
    flag: DWord;
    num_cpus: DWord;
    end;

  PTraceRecordHeader = ^TTraceRecordHeader;
  TTraceRecordHeader = packed record
    Len: Word;
    Major: Word;
    Minor: DWord;
    Dummy: DWord;
    end;

  PTraceInfo = ^TTraceInfo;
  TTraceInfo = record
    Pid, Tid, Addr: DWord;
    end;
var
  Fs: TFileStream;
  Hdr: PPerfHeader;
  HdrSize, DataSize: integer;
  i, Cpu, Idx: integer;
  Buf: array of Byte;
  Hook: PTraceRecordHeader;
  Info: PTraceInfo;
  Cnt: integer;
begin
  TotalHits := 0;
  ProcessHits := 0;
  BufferFull := false;
  Cnt := 0;

  Fs := TFileStream.Create(AFileName, fmOpenRead or fmShareDenyWrite);
  try
    //read header
    SetLength(Buf, SizeOf(TPerfHeader));
    Fs.ReadBuffer(Buf[0], Length(Buf));
    Hdr := @Buf[0];
    //analyze header
    if Hdr.nrm2 <> NRM2_SIGNATURE then
      raise Exception.Create('File corrupt: ' + AFileName);
    HdrSize := Hdr.header_size;
    CpuCount := Hdr.num_cpus;

    Fs.Position := 0;
    SetLength(Samples, 1024);

    //for each cpu section
    for i:=0 to CpuCount-1 do
      begin
      //section header
      SetLength(Buf, HdrSize);
      Fs.ReadBuffer(Buf[0], HdrSize);
      Hdr := @Buf[0];
      Cpu := Hdr.cpu_no;
      BufferFull := BufferFull or (Hdr.flag <> 0); 

      //read data
      DataSize := Hdr.data_size;
      SetLength(Buf, DataSize);
      if DataSize = 0 then Continue;
      Fs.ReadBuffer(Buf[0], DataSize);
      Idx := 0;

      //for each hook in data
      repeat
        Hook := @Buf[Idx];
        Hook.Len := Hook.Len and RAWFMT_HOOK_LEN_MASK;
        if Hook.Major = SFT_HOOK_TPROF_MAJOR then
          begin
          Info := @Buf[Idx + SizeOf(TTraceRecordHeader)];
          Inc(TotalHits);
          //store sample
          if Info.Pid = Pid then
            begin
            Inc(ProcessHits);

            if Cnt = Length(Samples) then SetLength(Samples, Length(Samples) * 2);
            Samples[Cnt].Thr := Info.Tid;
            Samples[Cnt].Addr := Info.Addr;
            Samples[Cnt].Cpu := Cpu;
            Samples[Cnt].Mode := 1 - (Hook.Minor and 1);

            Inc(Cnt);
            end;
          end;
        Inc(Idx, Hook.Len);
      until
        Idx >= Length(Buf);
      end;
  finally
    Fs.Free;
    SetLength(Samples, Cnt);
  end;
end;


procedure TPerformanceData.ResetCounters;
var
  i: integer;
begin
  CpuHits := nil;
  ModeHits := nil;
  ThrHits := nil;

  CpuThrHits := nil;
  ModeThrHits := nil;


  SetLength(CpuHits, CpuCount);
  SetLength(ModeHits, 2);
  SetLength(ThrHits, Threads.Count);

  SetLength(CpuThrHits, CpuCount, Threads.Count);
  SetLength(ModeThrHits, 2, Threads.Count);

  for i:=0 to High(HitsByModule) do FreeAndNil(HitsByModule[i]);
  HitsByModule := nil;
  SetLength(HitsByModule, Modules.Count);
end;


function SCompare(const S1, S2: TSample): integer;
begin
  if S1.Addr > S2.Addr then Result := 1
  else if S1.Addr < S2.Addr then Result := -1
  else if S1.Thr > S2.Thr then Result := 1
  else if S1.Thr < S2.Thr then Result := -1
  else Result := 0;
end;


procedure QuickSort(Arr: TSamples; First, Last: integer);
var
  I, J: integer;
  P, T: TSample;
begin
  repeat
    I := First;
    J := Last;
    P := Arr[(First + Last) shr 1];
    repeat
      while SCompare(Arr[I], P) < 0 do
        Inc(I);
      while SCompare(Arr[J], P) > 0 do
        Dec(J);
      if I <= J then
      begin
        T := Arr[I];
        Arr[I] := Arr[J];
        Arr[J] := T;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if First < J then QuickSort(Arr, First, J);
    First := I;
  until I >= Last;
end;


procedure TPerformanceData.AnalyzeSamples;
var
  i, Src, Dst: integer;
begin
  ResetCounters;
  if Samples = nil then Exit;

  //count hits by cpu, mode and thread
  for i:=0 to High(Samples) do
    with Samples[i] do
      begin
      if not Threads.Find(Thr, Integer(Thr)) then
        begin Samples := nil; Exit; end;

      Inc(CpuHits[Cpu]);
      Inc(ModeHits[Mode]);
      Inc(ThrHits[Thr]);

      Inc(CpuThrHits[Cpu, Thr]);
      Inc(ModeThrHits[Mode, Thr]);
      end;

  //group samples by addr/thread
  QuickSort(Samples, 0, High(Samples));

  Dst := 0;
  Samples[0].Count := 1;
  for Src:=1 to High(Samples) do
    if SCompare(Samples[Src], Samples[Dst]) = 0
      then
        begin
        Inc(Samples[Dst].Count);
        Samples[Dst].Mode := Samples[Dst].Mode or Samples[Src].Mode;
        end
      else
        begin
        Inc(Dst);
        Samples[Dst] := Samples[Src];
        Samples[Dst].Count := 1;
        end;

  SetLength(Samples, Dst+1);
end;


procedure TPerformanceData.LoadDebugInfo;
var
  i, M: integer;
  Location: TJclLocationInfo;
begin
  if Samples = nil then Exit;

  M := -1;

  for i:=0 to High(Samples) do
    begin
    if (M < 0) or not Modules[M].IsAddressInModule(Samples[i].Addr) then
      begin
      M := Modules.GetModuleIndexFromAddr(Samples[i].Addr);
      HitsByModule[M] := TModuleHits.Create(Threads.Count);
      HitsByModule[M].Name := Modules[M].Name;
      HitsByModule[M].ModuleInfo := Modules[M];
      //show progress
      if Assigned(FOnProcessModule) then
        FOnProcessModule(Self, Modules[M].Name, M / Modules.Count);
      end;

    //get debug info
    Modules[M].DebugInfoSource.GetLocationInfo(Ptr(Samples[i].Addr), Location);

    //count hits by module, unit, function, line
    HitsByModule[M].AddHits(Samples[i], Location);
    end;
end;


function TPerformanceData.IsEmpty: boolean;
begin
  Result := ProcessHits = 0;
end;


function TPerformanceData.FindSample(Addr: DWord): integer;
var
  H, I: integer;
begin
  Result := 0;
  H := High(Samples);

  while Result <= H do
    begin
    I := (Result + H) shr 1;
    if Samples[I].Addr < Addr then Result := I + 1 else  H := I - 1;
    end;
end;



end.

