//------------------------------------------------------------------------------
//The contents of this file are subject to the Mozilla Public License
//Version 1.1 (the "License"); you may not use this file except in compliance
//with the License. You may obtain a copy of the License at
//http://www.mozilla.org/MPL/ Software distributed under the License is
//distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
//or implied. See the License for the specific language governing rights and
//limitations under the License.
//
//The Original Code is RunApp.pas.
//
//The Initial Developer of the Original Code is Alex Shovkoplyas, VE3NEA.
//Portions created by Alex Shovkoplyas are
//Copyright (C) 2008 Alex Shovkoplyas. All Rights Reserved.
//------------------------------------------------------------------------------
unit RunApp;

interface

uses
  Windows, Messages, SysUtils, Classes, Psapi;


type
  TRunThread = class;


  TRunApplication = class
  private
    Thr: TRunThread;
    FNeedDebugEvents: boolean;
    FTerminated: boolean;
    FOnTerminate: TNotifyEvent;
    FOnDebugEvent: TNotifyEvent;

    function Run: boolean;
    procedure Cleanup;
    procedure GetDebugString;
    function ReadProcessString(Addr: PChar; Len: integer; Unicode: boolean): WideString;
  public
    //input
    ExePath: TFileName;
    CmdLine: string;
    ShowWindow: integer;
    //output
    ProcessInfo: TProcessInformation;
    ExitCode: DWord;
    DebugEvent: TDebugEvent;
    DebugString: WideString;

    constructor Create;
    destructor Destroy; override;

    //run and wait until terminated
    function RunSync: boolean;
    //run and notify when terminated
    function RunAsync(NeedDebugEvents: boolean = false): boolean;

    function IsRunning: boolean;
    function Stop: boolean;

    property OnTerminate: TNotifyEvent read FOnTerminate write FOnTerminate;
    property OnDebugEvent: TNotifyEvent read FOnDebugEvent write FOnDebugEvent;
  end;


  TRunThread = class(TThread)
  private
    procedure ProcessDebugEvent;
  protected
    Papa: TRunApplication;
    StartEventHandle: THandle;

    constructor Create(APapa: TRunApplication);
    procedure Execute; override;
    procedure AppTerminated;
    procedure DbgEvent;
  end;



implementation


{ TRunApplication }


//------------------------------------------------------------------------------
//                                system
//------------------------------------------------------------------------------
constructor TRunApplication.Create;
begin
  ShowWindow := SW_SHOWNORMAL;
  FTerminated := true;
end;


destructor TRunApplication.Destroy;
begin
  Stop;
  inherited;
end;






//------------------------------------------------------------------------------
//                                 run
//------------------------------------------------------------------------------
function TRunApplication.RunSync: boolean;
begin
  FNeedDebugEvents := false;
  Result := Run;
  if not Result then Exit;
  WaitForSingleObject(ProcessInfo.hProcess, INFINITE);
  Cleanup;
end;


function TRunApplication.RunAsync(NeedDebugEvents: boolean = false): boolean;
begin
  FNeedDebugEvents := NeedDebugEvents;
  Stop;
  Thr := TRunThread.Create(Self);
  Result := IsRunning;
  if not Result then Thr.WaitFor;
end;


function TRunApplication.Run: boolean;
var
  StartupInfo: TStartupInfo;
  CreationFlags: integer;
begin
  Result := false;
  if IsRunning then Exit;
  ExitCode := 0;

  ZeroMemory(@StartupInfo, SizeOf(StartupInfo));
  ZeroMemory(@ProcessInfo, SizeOf(ProcessInfo));

  with StartupInfo do
    begin
    cb := SizeOf(StartupInfo);
    dwFlags := STARTF_USESHOWWINDOW;
    wShowWindow := ShowWindow;
    end;

  if FNeedDebugEvents
    then CreationFlags := NORMAL_PRIORITY_CLASS or DEBUG_PROCESS or DEBUG_ONLY_THIS_PROCESS
    else CreationFlags := NORMAL_PRIORITY_CLASS;

  if not CreateProcess(PChar(ExePath), PChar(CmdLine), nil, nil, false,
    CreationFlags, nil, nil, StartupInfo, ProcessInfo)
      then Exit;

  FTerminated := false;
  Result := true;
end;


function TRunApplication.IsRunning: boolean;
begin
  Result := not FTerminated;
end;






//------------------------------------------------------------------------------
//                                 stop
//------------------------------------------------------------------------------
function TRunApplication.Stop: boolean;
var
  ExitCode: DWord;
begin
  if IsRunning then
    begin
    TerminateProcess(ProcessInfo.hProcess, 1);
    Thr.WaitFor;
    Cleanup;
    end;

  GetExitCodeProcess(ProcessInfo.hProcess, ExitCode);
  Result := ExitCode <> STILL_ACTIVE;
end;


procedure TRunApplication.Cleanup;
begin
  FTerminated := true;
  GetExitCodeProcess(ProcessInfo.hProcess, ExitCode);
  CloseHandle(ProcessInfo.hProcess);
  CloseHandle(ProcessInfo.hThread);
  ZeroMemory(@ProcessInfo, SizeOf(ProcessInfo));
end;






//------------------------------------------------------------------------------
//                                 thread
//------------------------------------------------------------------------------

{ TRunThread }

constructor TRunThread.Create(APapa: TRunApplication);
begin
  inherited Create(true);
  Papa := APapa;
  FreeOnTerminate := true;

  //wait until the app starts or fails
  StartEventHandle := CreateEvent(nil, false, false, nil);
  Resume;
  WaitForSingleObject(StartEventHandle, INFINITE);
  CloseHandle(StartEventHandle);
end;


procedure TRunThread.Execute;
var
  rc: DWord;
begin
  if not Papa.Run then Terminate;
  SetEvent(StartEventHandle);

  if Papa.FNeedDebugEvents
    then //wait for debug events
      begin
      while not Terminated do
        if WaitForDebugEvent(Papa.DebugEvent, 100) and not Terminated
          then ProcessDebugEvent;
      end

    else //wait for termination
      while not Terminated do
        begin
        rc := WaitForSingleObject(Papa.ProcessInfo.hProcess, 500);
        if (rc = WAIT_OBJECT_0) then Synchronize(AppTerminated);
        end;
end;




procedure TRunThread.ProcessDebugEvent;
var
  ContinueStatus: DWord;
begin
  ContinueStatus := DBG_CONTINUE;

  case Papa.DebugEvent.dwDebugEventCode of
    CREATE_PROCESS_DEBUG_EVENT:
      begin
      Synchronize(DbgEvent);
      CloseHandle(Papa.DebugEvent.CreateProcessInfo.hFile);
      end;

    EXIT_PROCESS_DEBUG_EVENT:
      Synchronize(AppTerminated);

    CREATE_THREAD_DEBUG_EVENT:
      Synchronize(DbgEvent);

    LOAD_DLL_DEBUG_EVENT:
      begin
      Synchronize(DbgEvent);
      CloseHandle(Papa.DebugEvent.LoadDll.hFile);
      end;

    OUTPUT_DEBUG_STRING_EVENT:
      begin
      Papa.GetDebugString;
      Synchronize(DbgEvent);
      end;

    EXCEPTION_DEBUG_EVENT:
      if (Papa.DebugEvent.Exception.dwFirstChance = 0)
        or (Papa.DebugEvent.Exception.ExceptionRecord.ExceptionFlags <> 0)
       then ContinueStatus := DBG_EXCEPTION_NOT_HANDLED;
    end;

  ContinueDebugEvent(Papa.DebugEvent.dwProcessId, Papa.DebugEvent.dwThreadId, ContinueStatus);
end;    


procedure TRunThread.AppTerminated;
begin
  Papa.Cleanup;
  Papa.Thr := nil;
  if Assigned(Papa.FOnTerminate) then Papa.FOnTerminate(Papa);
  Terminate;
end;


procedure TRunThread.DbgEvent;
begin
  if Assigned(Papa.FOnDebugEvent) then Papa.FOnDebugEvent(Papa);
end;


function TRunApplication.ReadProcessString(Addr: PChar; Len: integer; Unicode: boolean): WideString;
var
  S: string;
  Cnt: DWord;
begin
  if Unicode
    then
      begin
      SetLength(Result, Len);
      ReadProcessMemory(ProcessInfo.hProcess,  Addr, @Result[1], Len * SizeOf(WideChar), Cnt);
      end
    else
      begin
      SetLength(S, Len);
      ReadProcessMemory(ProcessInfo.hProcess, Addr, @S[1], Len, Cnt);
      Result := S;
      end;
end;


procedure TRunApplication.GetDebugString;
begin
  with DebugEvent.DebugString do
    DebugString := ReadProcessString(@lpDebugStringData^, nDebugStringLength-1, fUnicode <> 0);
end;



end.

