//------------------------------------------------------------------------------
//The contents of this file are subject to the Mozilla Public License
//Version 1.1 (the "License"); you may not use this file except in compliance
//with the License. You may obtain a copy of the License at
//http://www.mozilla.org/MPL/ Software distributed under the License is
//distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
//or implied. See the License for the specific language governing rights and
//limitations under the License.
//
//The Original Code is Samplr.pas.
//
//The Initial Developer of the Original Code is Alex Shovkoplyas, VE3NEA.
//Portions created by Alex Shovkoplyas are
//Copyright (C) 2008 Alex Shovkoplyas. All Rights Reserved.
//------------------------------------------------------------------------------
unit Samplr;

interface

uses
  Windows, SysUtils, Classes, Dialogs, ShellApi, Registry, PathUtil, PerfUtil;


type
  TSampler = class
  public
    EventNames: TStringList;

    RawFileName: TFileName;
    BufSize: integer;
    SamplingRate: integer;
    EventBased: boolean;
    EventId: integer;
    EventCount: integer;

    constructor Create;
    destructor Destroy; override;
    procedure ListEvents;

    function CheckInitialized: boolean;
    procedure Open;
    procedure Close;
    procedure Start;
    procedure Stop;
    function GetEventDescription(EventId: integer): string;
    end;



implementation

{ TSampler }

constructor TSampler.Create;
begin
  SamplingRate := IBMPERF_TPROF_TICK_RATE;
  BufSize := PI_TRACE_BUF_DEFAULT;
  RawFileName := ChangeFileExt(ParamStr(0), '.nrm2');
  EventCount := 1000000;

  EventNames := TStringList.Create;
  if IsPerfUtilLoaded then ListEvents;
end;


destructor TSampler.Destroy;
begin
  EventNames.Free;
  inherited;
end;  


function TSampler.CheckInitialized: boolean;
const
  Url = 'http://perfinsp.sourceforge.net/';
begin
  Result := IsPerfUtilLoaded;

  if not Result then
    if idYes = MessageDlg('Performance Inspector is not installed.'#13 +
               'Open the download page?', mtError, [mbYes, mbNo], 0)

      then ShellExecute(GetDesktopWindow, 'open', PChar(Url), '', '', SW_SHOWNORMAL);
end;


procedure TSampler.Open;
var
  //for debugging
  rc: integer;
begin
  if not IsPerfUtilLoaded then Exit;

  rc := TraceInit(BufSize);

  if EventBased
    then rc := TraceSetTprofEvent(ALL_CPUS, EventId, EventCount, PRIVILEGE_LEVEL_ALL)
    else rc := SetProfilerRate(SamplingRate);

  rc := TraceEnable($10);
end;


procedure TSampler.Close;
begin
  if not IsPerfUtilLoaded then Exit;

  TraceOff;
  try
    if TraceWriteBufferToFile(PChar(RawFileName)) <> 0 then
      Exception.Create('TraceWriteBufferToFile failed');
  finally
    TraceTerminate;
  end;
end;


procedure TSampler.Start;
begin
  if not IsPerfUtilLoaded then Exit;

  if IsTraceOn = 0 then TraceOn else TraceResume;
end;


procedure TSampler.Stop;
begin
  if not IsPerfUtilLoaded then Exit;

  TraceSuspend;
end;


procedure TSampler.ListEvents;
var
  Cnt, i: integer;
  IdPtr: PInteger;
  Info: PPerfCtrEventInfo;
begin
  Cnt := GetPerfCounterEventIdList(IdPtr);
  if Cnt = 0 then Exit;

  for i:=0 to Cnt-1 do
    begin
    Info := GetPerfCounterEventInfoFromId(IdPtr^);
    EventNames.AddObject(Info.name, Ptr(Info.event_id));
    Inc(IdPtr);
    end;
end;


function TSampler.GetEventDescription(EventId: integer): string;
var
  Info: PPerfCtrEventInfo;
begin
  Result := '';
  if not IsPerfUtilLoaded then Exit;

  Info := GetPerfCounterEventInfoFromId(EventId);
  if Info = nil then Exit;

  Result := Info.description;
end;



end.

