//------------------------------------------------------------------------------
//The contents of this file are subject to the Mozilla Public License
//Version 1.1 (the "License"); you may not use this file except in compliance
//with the License. You may obtain a copy of the License at
//http://www.mozilla.org/MPL/ Software distributed under the License is
//distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express
//or implied. See the License for the specific language governing rights and
//limitations under the License.
//
//The Original Code is Sett.pas.
//
//The Initial Developer of the Original Code is Alex Shovkoplyas, VE3NEA.
//Portions created by Alex Shovkoplyas are
//Copyright (C) 2008 Alex Shovkoplyas. All Rights Reserved.
//------------------------------------------------------------------------------
unit Sett;

interface

uses
  SysUtils, Classes, IniFiles, Menus, Math, PathUtil;

type
  TSettings = record
    //app start/stop
    AppName: TFileName;
    StartDelayed: boolean;
    StartPaused: boolean;
    AutoStop: boolean;
    WarmUpSeconds: integer;
    SampingSeconds: integer;

    //sampling
    BufSize: integer;
    SamplingRate: integer;
    EventBased: boolean;
    EventId: integer;
    EventCount: integer;

    //symbols/src
    SymbolsPath: TFileName;
    CanDownload: boolean;
    BorTool: integer;
    end;


  TMruList = class
  private
    FIni: TIniFile;
    FIniSections: TStringList;
    FItems: array[0..7] of TSettings;
    FCount: integer;
    FDefault: TSettings;

    procedure LoadItem(ItemNo: integer);
    function GetItem(Index: integer): TSettings;
    procedure SaveItem(ItemNo: integer);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Load(AIniFile: TFileName);
    procedure Save(AIniFile: TFileName);
    procedure Add(const ASettings: TSettings);
    procedure ToMenu(AfterItem: TMenuItem);
    property Items[Index: integer]: TSettings read GetItem; default;
  end;


implementation

{ TMruList }

constructor TMruList.Create;
begin
  FIniSections := TStringList.Create;

  //define default settings

  FDefault.WarmUpSeconds := 3;
  FDefault.SampingSeconds := 10;
  FDefault.SymbolsPath := WinDir + '\Symbols';
  FDefault.BufSize := 3 shl 20; //3 Mb
  FDefault.SamplingRate := 1024;
  FDefault.EventCount := 100000;
end;


destructor TMruList.Destroy;
begin
  FIniSections.Free;
  inherited;
end;


procedure TMruList.Add(const ASettings: TSettings);
var
  i, Idx: integer;
begin
  //item to delete
  Idx := Min(High(FItems), FCount);
  for i:=0 to FCount-1 do
    if CompareText(FItems[i].AppName, ASettings.AppName) = 0 then
      begin Idx := i; Break; end;

  //delete
  for i:=Idx downto 1 do FItems[i] := FItems[i-1];
  
  //add new
  FItems[0] := ASettings;
  if Idx = Min(High(FItems), FCount) then
    FCount := Min(Length(FItems), FCount + 1);
end;


function TMruList.GetItem(Index: integer): TSettings;
begin
  if (Index >= 0) and (Index < FCount)
    then Result := FItems[Index]
    else Result := FDefault;
end;


procedure TMruList.Load(AIniFile: TFileName);
var
  i: integer;
begin
  FIni := TIniFile.Create(AIniFile);
  try
    FIni.ReadSections(FIniSections);
    FCount := FIniSections.Count;
    for i:=0 to FCount-1 do LoadItem(i);
  finally
    FreeAndNil(FIni);
  end;
end;


procedure TMruList.LoadItem(ItemNo: integer);
var
  Sect: string;
begin
  Sect := Format('MRU_%d', [ItemNo]);

  FItems[ItemNo].AppName := FIni.ReadString(Sect, 'AppName', FDefault.AppName);
  FItems[ItemNo].StartDelayed := FIni.ReadBool(Sect, 'StartDelayed', FDefault.StartDelayed);
  FItems[ItemNo].StartPaused := FIni.ReadBool(Sect, 'StartPaused', FDefault.StartPaused);
  FItems[ItemNo].AutoStop := FIni.ReadBool(Sect, 'AutoStop', FDefault.AutoStop);
  FItems[ItemNo].WarmUpSeconds := FIni.ReadInteger(Sect, 'WarmUpSeconds', FDefault.WarmUpSeconds);
  FItems[ItemNo].SampingSeconds := FIni.ReadInteger(Sect, 'SampingSeconds', FDefault.SampingSeconds);
  FItems[ItemNo].BufSize := FIni.ReadInteger(Sect, 'BufSize', FDefault.BufSize);
  FItems[ItemNo].SamplingRate := FIni.ReadInteger(Sect, 'SamplingRate', FDefault.SamplingRate);
  FItems[ItemNo].EventBased := FIni.ReadBool(Sect, 'EventBased', FDefault.EventBased);
  FItems[ItemNo].EventId := FIni.ReadInteger(Sect, 'EventId', FDefault.EventId);
  FItems[ItemNo].EventCount := FIni.ReadInteger(Sect, 'EventCount', FDefault.EventCount);
  FItems[ItemNo].SymbolsPath := FIni.ReadString(Sect, 'SymbolsPath', FDefault.SymbolsPath);
  FItems[ItemNo].CanDownload := FIni.ReadBool(Sect, 'CanDownload', FDefault.CanDownload);
  FItems[ItemNo].BorTool := FIni.ReadInteger(Sect, 'BorTool', FDefault.BorTool);
end;


procedure TMruList.Save(AIniFile: TFileName);
var
  i: integer;
begin
  FIni := TIniFile.Create(AIniFile);
  try
    FIni.ReadSections(FIniSections);
    for i:=0 to FIniSections.Count-1 do FIni.EraseSection(FIniSections[i]);
    for i:=0 to FCount-1 do SaveItem(i);
  finally
    FreeAndNil(FIni);
  end;
end;


procedure TMruList.SaveItem(ItemNo: integer);
var
  Sect: string;
begin
  Sect := Format('MRU_%d', [ItemNo]);

  FIni.WriteString(Sect, 'AppName', FItems[ItemNo].AppName);
  FIni.WriteBool(Sect, 'StartDelayed', FItems[ItemNo].StartDelayed);
  FIni.WriteBool(Sect, 'StartPaused', FItems[ItemNo].StartPaused);
  FIni.WriteBool(Sect, 'AutoStop', FItems[ItemNo].AutoStop);
  FIni.WriteInteger(Sect, 'WarmUpSeconds', FItems[ItemNo].WarmUpSeconds);
  FIni.WriteInteger(Sect, 'SampingSeconds', FItems[ItemNo].SampingSeconds);
  FIni.WriteInteger(Sect, 'BufSize', FItems[ItemNo].BufSize);
  FIni.WriteInteger(Sect, 'SamplingRate', FItems[ItemNo].SamplingRate);
  FIni.WriteBool(Sect, 'EventBased', FItems[ItemNo].EventBased);
  FIni.WriteInteger(Sect, 'EventId', FItems[ItemNo].EventId);
  FIni.WriteInteger(Sect, 'EventCount', FItems[ItemNo].EventCount);
  FIni.WriteString(Sect, 'SymbolsPath', FItems[ItemNo].SymbolsPath);
  FIni.WriteBool(Sect, 'CanDownload', FItems[ItemNo].CanDownload);
  FIni.WriteInteger(Sect, 'BorTool', FItems[ItemNo].BorTool);
end;


procedure TMruList.ToMenu(AfterItem: TMenuItem);
var
  Par, Mnu: TMenuItem;
  Idx: integer;
  i: integer;
begin
  Par := AfterItem.Parent;
  Idx := Par.IndexOf(AfterItem) + 1;

  //clear menu
  while (Idx < Par.Count) and not Par[Idx].IsLine do Par[Idx].Free;

  //add mru items
  for i:=FCount-1 downto 0 do
    begin
    Mnu := TMenuItem.Create(Par);
    Mnu.Caption := FItems[i].AppName;
    Mnu.Tag := i;
    Mnu.OnClick := AfterItem.OnClick;
    Par.Insert(Idx, Mnu);
    end;

  //hide "No MRU"
  AfterItem.Visible := FCount = 0;
end;




end.

